from __future__ import annotations

from abc import ABC, abstractclassmethod
from functools import cached_property
from typing import Callable, TYPE_CHECKING, Type, TypeVar, Any, Protocol
from functools import cached_property
from .exceptions import SerizlierNotFound

if TYPE_CHECKING:
    from .sql import SQLEnvironment


class StringRepresentation(Protocol):
    def __str__(self) -> str:
        ...


NO_NAME_PARAM_NAME = "No Name"


class AdapterSerializer:
    def serialize_params(self, values: dict[str, Any]) -> dict[str, Any]:
        return {param: self.serialize_param(value) for param, value in values.items()}

    def serialize_param(self, value: Any, name: str = NO_NAME_PARAM_NAME) -> Any:
        match value:
            case int() | float():
                _value = str(value)
            case str():
                _value = self.serialize_string(value)
            case _:
                _value = self.serialize_other_object(value)

        return _value

    @staticmethod
    def serialize_number(value: int | float) -> str:
        return str(value)

    @staticmethod
    def serialize_string(value: str | StringRepresentation) -> str:
        return f"'{value}'"

    def serialize_other_object(self, value: Any, name: str = NO_NAME_PARAM_NAME) -> str:
        if "__str__" in vars(type(value)):
            _value = self.serialize_string(value)
        else:
            raise SerizlierNotFound(name, type(value))

        return _value


class Adapter(ABC):
    serializer: AdapterSerializer = AdapterSerializer()

    def __init__(
        self,
        sql: SQLEnvironment,
        connection_factory: Callable,
    ) -> None:
        self._sql = sql
        self._connection_factory = connection_factory

    @abstractclassmethod
    def execute(self) -> list[dict[str, Any]]:
        ...

    @cached_property
    def connection(self) -> Any:
        return self._connection_factory()

    @cached_property
    def query(self) -> str:
        return self._sql._query.render(params=self.serialized_params, envs=self.envs)

    @cached_property
    def envs(self) -> dict[str, Any]:
        return {
            **self.adapter_evns,
            **self._sql._envs,
        }

    @property
    def adapter_evns(self) -> dict[str, Any]:
        return {
            "cast": self.serializer.serialize_param,
        }

    @cached_property
    def serialized_params(self) -> dict[str, Any]:
        return self.serializer.serialize_params(self._sql._params)


A = TypeVar("A", bound=Adapter)


class AdapterFactory(ABC):
    adapter_factory: Type[A]

    def __init__(self, connection_factory: Callable) -> None:
        super().__init__()
        self._connection_factory = connection_factory

    @cached_property
    def connection(self) -> Any:
        return self._connection_factory()

    def __get__(
        self,
        sql: SQLEnvironment,
        enviroment: Type[SQLEnvironment] | None = None,
    ) -> A:
        return self.adapter_factory(sql, self._connection_factory)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sqle', 'sqle.contrib']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0']

setup_kwargs = {
    'name': 'sqle',
    'version': '0.0.0a0',
    'description': '',
    'long_description': '# TSQL\n\nPython package designed to execute pure sql queries.\n\n| :exclamation:  This is alpha version    |\n|-----------------------------------------|\n',
    'author': 'acrius',
    'author_email': 'acrius@mail.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from collections.abc import Mapping
from inspect import isclass, stack
from os.path import isabs
from pathlib import Path
from typing import Callable, Type, TypeVar, Any

from jinja2 import Environment
from typing_extensions import get_type_hints

from .adapter import AdapterFactory
from .contrib.class_property import classproperty
from .exceptions import TSQLValueError
from .query import Query
from .settings import SQL_ENVIRONMENT_DEFAULT_INSTANCE_NAME

T = TypeVar("T", bound="SQLEnvironment")


class SQLEnvironment:
    def __init__(self, query: Query | str) -> None:
        self._query = query if isinstance(query, Query) else Query(query)
        self._envs = {}
        self._params = {}

    def with_params(self, **params) -> SQLEnvironment:
        return self.copy(with_params=params)

    def with_envs(self, **envs) -> SQLEnvironment:
        return self.copy(with_envs=envs)

    def copy(
        self: T,
        with_envs: dict | None = None,
        with_params: dict | None = None,
    ) -> T:
        sql = self.__class__(self._query)

        if with_envs:
            if not isinstance(with_envs, Mapping):
                raise TSQLValueError(f"Envs must be `Mapping`, got {type(with_envs)}.")

            sql._extend_envs(**with_envs)

        if with_params:
            if not isinstance(with_params, Mapping):
                raise TSQLValueError(
                    f"Params nust be `Mapping`, get {type(with_params)}."
                )

            sql._extends_params(**with_params)

        return sql

    def _extend_envs(self, **envs: Any):
        self._envs = {
            **self._envs,
            **envs,
        }

    def _extends_params(self, **params: Any):
        self._params = {
            **self._params,
            **params,
        }

    # Factory

    @classmethod
    def create_instance(
        cls,
        name: str = SQL_ENVIRONMENT_DEFAULT_INSTANCE_NAME,
        **connection_factories: Callable,
    ) -> Type[SQLEnvironment]:
        sql_environment_instance = type(name, cls.__bases__, dict(cls.__dict__))

        for adapter_name, adapter_factory in cls.adapters.items():
            connection_factory = connection_factories.get(adapter_name)

            if not connection_factory and not cls.get_default_adapter(adapter_name):
                raise TSQLValueError(
                    f"Pass the connection factory or set the default adapter for {adapter_name}({adapter_factory}).",
                )

            if connection_factory:
                setattr(
                    sql_environment_instance,
                    adapter_name,
                    adapter_factory(connection_factory),
                )

        return sql_environment_instance

    @classmethod
    def get_default_adapter(cls, adapter_name: str) -> Adapter | None:
        return getattr(cls, adapter_name, None)

    @classproperty
    def adapters(cls) -> dict[str, type[AdapterFactory]]:
        return {
            field: type_
            for field, type_ in get_type_hints(cls).items()
            if isclass(type_) and issubclass(type_, AdapterFactory)
        }



from __future__ import annotations

from inspect import stack
from os.path import isabs
from pathlib import Path
from typing import Type, TypeVar

from jinja2 import Environment, FileSystemLoader
from .contrib.text import clear_text

from .exceptions import TSQLValueError

T = TypeVar("T", bound="Query")


class Query:
    TEXT_OR_PATH_MUST_BE_PASSER_ERROR = TSQLValueError(
        f"A text or path must be passed to the Query."
    )

    def __init__(
        self,
        text: str | None = None,
        path: str | None = None,
        template_environment: Environment | None = None,
    ) -> Query:
        if not text and not path:
            raise self.TEXT_OR_PATH_MUST_BE_PASSER_ERROR

        self._text = text
        self._path = path
        self._calling_path = self._get_calling_path()
        self._template_environment = (
            template_environment or self.create_template_environment()
        )

    def create_template_environment(self) -> Environment:
        return Environment(
            loader=FileSystemLoader(self._calling_path),
            trim_blocks=True,
            lstrip_blocks=True,
            keep_trailing_newline=True,
        )

    @classmethod
    def from_file(
        cls: Type[T],
        path: str | Path,
        template_environment: Environment | None = None,
    ) -> T:
        return cls(path=path, template_environment=template_environment)

    @staticmethod
    def _get_calling_path() -> Path:
        calling_context = next(
            context for context in stack() if context.filename != __file__
        )
        return Path(calling_context.filename).parent

    def render(self, envs: dict | None = None, params: dict | None = None) -> str:
        envs = envs or {}
        rendered_template = self.render_template()

        if params:
            rendered_text = rendered_template.format(**params)
        else:
            rendered_text = rendered_template

        return clear_text(rendered_text)

    def render_template(self, **envs):
        if self._text:
            template = self._template_environment.from_string(self._text)
        elif self._path:
            template = self._template_environment.get_template(self._path)
        else:
            raise self.TEXT_OR_PATH_MUST_BE_PASSER_ERROR

        return template.render(**envs)



from __future__ import annotations


class TSQLException(Exception):
    ...


class TSQLValueError(TSQLException, ValueError):
    ...


class SerizlierNotFound(TSQLException, ValueError):
    _message = "Serializer not found for {param_name}({param_type})"

    def __init__(self, param_name, param_type) -> None:
        message = self._message.format(param_name=param_name, param_type=param_type)
        super().__init__(message)
